/*-----------------------------------------------------------------------

File  : can_infstate.c

Author: Stephan Schulz

Contents

  Functions for representing and analysing the state of an inference
  process.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Sun Feb 28 19:06:13 MET 1999
    New

-----------------------------------------------------------------------*/

#include "can_infstate.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

#ifndef NDEBUG

/*-----------------------------------------------------------------------
//
// Function: verify_clause_term_assoc()
//
//   For each term in clause check if term is associated with clause.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static bool verify_clause_term_assoc(InfState_p state, CompClause_p
				     clause, char* called)
{
   int i;
   PTree_p *root;
   Term_p dummy;
   Term_p term;

   assert(clause->lit_terms);
   
   for(i=0; i<2*clause->literal_no; i++)
   {
      term = clause->lit_terms[i];
      dummy = PDArrayElementP(state->term_clause_assoc,
			      term->entry_no);
      root = (PTree_p*)
	 &(state->term_clause_assoc->array
	   [term->entry_no].p_val);
      if(!PTreeFind(root, clause))
      {
	 fprintf(GlobalOut, 
		 "# verify_clause_term_assoc(%s):"
		 " %ld is not associated with %ld\n",
		 called, clause->ident, term->entry_no);
      }
   }
   return true;
}
#endif


/*-----------------------------------------------------------------------
//
// Function: print_example_conditional()
//
//   Print a clause and appropriate information in the correct format
//   for an example file (if clause is watched).
//
// Global Variables: GlobalOut
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void print_example_conditional(CompClause_p clause, InfState_p
				      state)
{
   if(ClauseQueryProp(clause, CPIsWatched))
   {
      fprintf(GlobalOut, "%6ld:", clause->ident);      
      if(state->print_clause_stats)
      {
	 ClauseStatsPrint(GlobalOut, &(clause->stats));
      }
      fprintf(GlobalOut, ":");
      CompClausePrint(GlobalOut, clause, state->terms, true);
      fprintf(GlobalOut, "\n");
   }
}


/*-----------------------------------------------------------------------
//
// Function: print_example_tree()
//
//   For all clauses in tree, conditionally print them as examples. 
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void print_example_tree(PTree_p tree, InfState_p state)
{
   PStack_p stack;
   PTree_p  node;

   stack = PTreeTraverseInit(tree);
   
   while((node = PTreeTraverseNext(stack)))
   {
      print_example_conditional(node->key, state);
   }
   PTreeTraverseExit(stack);
}


/*-----------------------------------------------------------------------
//
// Function: find_associated_clauses()
//
//   Find all clauses associated with term and its superterms and add
//   them to the tree.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

static void find_associated_clauses(InfState_p state, Term_p term,
				    PTree_p *store)
{
   PStack_p ptree_stack, term_stack;
   PTree_p node;

   assert(term);
   
   term_stack = PStackAlloc();
   PStackPushP(term_stack, term);

   while(!PStackEmpty(term_stack))
   {
      term = PStackPopP(term_stack);
      assert(term);

      PTreeInsertTree(store, PDArrayElementP(state->term_clause_assoc,
					     term->entry_no));
      ptree_stack = PTreeTraverseInit(term->superterms);
      while((node = PTreeTraverseNext(ptree_stack)))
      {
         PStackPushP(term_stack, node->key);
      }  
      PTreeTraverseExit(ptree_stack);
   }
   PStackFree(term_stack);   
}


/*-----------------------------------------------------------------------
//
// Function: print_parent_tree()
//
//   Print the list of clause identifiers represented by parents.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void print_parent_tree(FILE* out, PTree_p parents)
{
   CompClause_p parent;
   PTree_p      cell;
   PStack_p     stack;
   char*        prefix = "";

   fprintf(out, "(");      
   stack = PTreeTraverseInit(parents);
   while((cell = PTreeTraverseNext(stack)))
   {
      parent = cell->key;
      assert(parent);
      fprintf(out, "%s%ld", prefix, parent->ident);
      prefix = ",";      
   }
   PTreeTraverseExit(stack);
   fprintf(out, ")");   
}


/*-----------------------------------------------------------------------
//
// Function: parse_parent_list()
//
//   Parse a list of parents for a clause and insert them into ethe
//   tree.
//
// Global Variables: -
//
// Side Effects    : Reads input
//
/----------------------------------------------------------------------*/

static long parse_parent_list(Scanner_p in, InfState_p state, PTree_p
			      *parents)
{
   long         clause_id;
   CompClause_p compact; 
   long res = 0;

   AcceptInpTok(in, OpenBracket);
   
   while(!TestInpTok(in,CloseBracket))
   {
      CheckInpTok(in, PosInt);
      clause_id = AktToken(in)->numval;

      compact = PDArrayElementP(state->clause_store, clause_id);
      if(!compact)
      {
	 AktTokenError(in, "Unknown clause in parent list", false);
      }
      PTreeStore(parents,compact);
      assert(clause_id == compact->ident);
      PTreeStore(parents,compact);
      NextToken(in);
      if(!TestInpTok(in,CloseBracket))
      {
	 AcceptInpTok(in, Comma);
      }
      res++;
   }
   AcceptInpTok(in, CloseBracket);
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: store_rewrite_parent()
//
//   For all clauses associated with term, add clause as a
//   rewrite-parent. 
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

static void store_rewrite_parent(InfState_p state, Term_p term,
				 CompClause_p clause)
{   
   PStack_p     clause_stack;
   CompClause_p handle;
   PTree_p      node;

   assert(clause);
   assert(clause->literal_no == 1);
   assert(state);
   assert(term);
   
   clause_stack =
      PTreeTraverseInit(PDArrayElementP(state->term_clause_assoc,
					term->entry_no));
   
   while((node = PTreeTraverseNext(clause_stack)))
   {
      handle = node->key;
      PTreeStore(&(handle->s_parents), clause);
   }
   PTreeTraverseExit(clause_stack);
}

/*-----------------------------------------------------------------------
//
// Function: store_rewrite_parents()
//
//   For all clauses associated with term and its superterms, add
//   clause as a rewrite-parent.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rek_store_rewrite_parent(InfState_p state, Term_p term,
				     CompClause_p clause)
{
   PStack_p term_stack, ptree_stack;
   PTree_p  node;

   assert(clause);
   assert(clause->literal_no == 1);
   assert(state);
   assert(term);

   term_stack = PStackAlloc();
   PStackPushP(term_stack, term);

   while(!PStackEmpty(term_stack))
   {
      term = PStackPopP(term_stack);
      assert(term);

      store_rewrite_parent(state, term, clause);
      
      ptree_stack = PTreeTraverseInit(term->superterms);
      while((node = PTreeTraverseNext(ptree_stack)))
      {
	 PStackPushP(term_stack, node->key);
      }	 
      PTreeTraverseExit(ptree_stack);
   }
   PStackFree(term_stack);   
}

/*-----------------------------------------------------------------------
//
// Function: change_term_clause_association()
//
//   Remove all clauses associated with term and associate them with
//   repl. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void change_term_clause_association(InfState_p state, Term_p
					   term, Term_p repl)
{
   PTree_p *root, source;
   long t_index, r_index;

   t_index = term->entry_no;
   r_index = repl->entry_no;
   
   assert(t_index<=state->term_clause_assoc->size);
   assert(r_index<=state->term_clause_assoc->size);
  
   root = (PTree_p*)
      &(state->term_clause_assoc->array[r_index].p_val);
   source = state->term_clause_assoc->array[t_index].p_val;
   PTreeMerge(root, source);
   PDArrayAssignP(state->term_clause_assoc, term->entry_no,
		  NULL);
}


/*-----------------------------------------------------------------------
//
// Function: parse_term_list()
//
//   Parse a list of terms and push them on the stack.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

static void parse_term_list(Scanner_p in, TB_p bank, PStack_p stack)
{
   Term_p term;
   
   AcceptInpTok(in, OpenSquare);

   while(!TestInpTok(in, CloseSquare))
   {
      term = TBTermParse(in, bank);
      PStackPushP(stack, term);
   }
   AcceptInpTok(in, CloseSquare); 
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: InfStateAlloc()
//
//   Allocate an empty inference process state.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

InfState_p InfStateAlloc()
{
   InfState_p handle = InfStateCellAlloc();

   handle->terms             = TBAlloc(TPRestricted,SigAlloc());
   handle->clause_store      = PDArrayAlloc(100000,100000);
   handle->term_clause_assoc = PDArrayAlloc(100000,100000);
   handle->watch_clauses     = NULL;
   handle->print_clause_evals = false;
   handle->print_clause_mods  = false;
   handle->print_clause_stats = true;
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: InfStateFree()
//
//   Return all the memory taken up by a InfState.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void InfStateFree(InfState_p state)
{
   long i;
   CompClause_p clause;
   PTree_p tree;
   
   for(i=0; i<state->clause_store->size; i++)
   {
      clause = PDArrayElementP(state->clause_store,i);
      if(clause)
      {
	 CompClauseFree(clause, state->terms);
	 PDArrayAssignP(state->clause_store, i, NULL);
      }      
   }
   for(i=0; i<state->term_clause_assoc->size; i++)
   {
      tree = PDArrayElementP(state->term_clause_assoc,i);
      if(tree)
      {
	 PTreeFree(tree);
	 PDArrayAssignP(state->term_clause_assoc, i, NULL);
      }      
   }
   PDArrayFree(state->clause_store);
   PDArrayFree(state->term_clause_assoc);
   
   SigFree(state->terms->sig);
   state->terms->sig = NULL;
   TBFree(state->terms);
   ClauseStatTreeFree(state->watch_clauses);

   InfStateCellFree(state);
}


/*-----------------------------------------------------------------------
//
// Function: InfStateClausesDelProp()
//
//   Delete the property from all clauses in state.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void InfStateClausesDelProp(InfState_p state, ClauseProperties props)
{
   long i;
   CompClause_p clause;
   
   for(i=0; i<state->clause_store->size; i++)
   {
      clause = PDArrayElementP(state->clause_store,i);
      if(clause)
      {
	 ClauseDelProp(clause, props);
      }      
   }   
}


/*-----------------------------------------------------------------------
//
// Function: InfStateClausesSetProp()
//
//   Set the property from all clauses in state.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void InfStateClausesSetProp(InfState_p state, ClauseProperties props)
{
   long i;
   CompClause_p clause;
   
   for(i=0; i<state->clause_store->size; i++)
   {
      clause = PDArrayElementP(state->clause_store,i);
      if(clause)
      {
	 ClauseSetProp(clause, props);
      }      
   }   
}


/*-----------------------------------------------------------------------
//
// Function: InfStateAddTermAssoc()
//
//   Given a state and a clause, add the term->clause associations to
//   the state.
//
// Global Variables: -
//
// Side Effects    : Changes state
//
/----------------------------------------------------------------------*/

void InfStateAddTermAssoc(InfState_p state, CompClause_p clause)
{
   int i;   
   PTree_p *root;
   Term_p dummy;

   assert(clause->lit_terms);
   
   for(i=0; i<2*clause->literal_no; i++)
   {
      dummy = PDArrayElementP(state->term_clause_assoc,
			      clause->lit_terms[i]->entry_no);
      root = (PTree_p*)
	 &(state->term_clause_assoc->array
	   [clause->lit_terms[i]->entry_no].p_val);
      PTreeStore(root, clause);
   }
}


/*-----------------------------------------------------------------------
//
// Function: InfStateRemoveTermAssoc()
//
//   Given a state and a clause, remove the association of the terms
//   with the clause.
//
// Global Variables: - 
//
// Side Effects    : Changes state
//
/----------------------------------------------------------------------*/

void InfStateRemoveTermAssoc(InfState_p state, CompClause_p clause)
{
   int i;   
   PTree_p *root;
   Term_p dummy;

   assert(clause->lit_terms);
   
   for(i=0; i<2*clause->literal_no; i++)
   {
      dummy = PDArrayElementP(state->term_clause_assoc,
			      clause->lit_terms[i]->entry_no);
      root = (PTree_p*)
	 &(state->term_clause_assoc->array
	   [clause->lit_terms[i]->entry_no].p_val);
      PTreeDeleteEntry(root, clause);
   }
}


/*-----------------------------------------------------------------------
//
// Function: InfStateCCParse()
//
//   Parse a CC (Clause creation) inference step into state.
//
// Global Variables: -
//
// Side Effects    : Changes state, reads input.
//
/----------------------------------------------------------------------*/

void InfStateCCParse(Scanner_p in, InfState_p state)
{
   Clause_p      clause;
   CompClause_p  compact;
   PTree_p       parents = NULL;
   long          clause_id;
   NumTree_p     cell;
   ClauseStats_p stats;
   
   AcceptInpId(in, "CC");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   if(PDArrayElementP(state->clause_store, clause_id))
   {
      AktTokenError(in, "Clause to be created already exists", false);
   }
   /*  printf("CC %ld\n", clause_id); */
   NextToken(in);
   AcceptInpTok(in, Ident);
   parse_parent_list(in, state, &parents);
   clause = ClauseParse(in, state->terms);
   clause->ident = clause_id;
      
   compact = CompactifyClause(clause);
   compact->g_parents = parents;
   
   if((cell = NumTreeFind(&(state->watch_clauses), clause_id)))
   {      
      ClauseSetProp(compact, CPIsWatched);
      stats = cell->val1.p_val;
      compact->stats = *stats;
   }
   PDArrayAssignP(state->clause_store, clause_id, compact);   
   InfStateAddTermAssoc(state, compact);
   if(state->print_clause_mods)
   {
      print_example_conditional(compact, state);
   }
   assert(verify_clause_term_assoc(state, compact, "CC"));
   }


/*-----------------------------------------------------------------------
//
// Function: InfStateCRParse()
//
//   Parse a CR (Clause renaming) inference step into state.
//
// Global Variables: -
//
// Side Effects    : Changes state, reads input.
//
/----------------------------------------------------------------------*/

void InfStateCRParse(Scanner_p in, InfState_p state)
{
   Clause_p clause;
   CompClause_p compact;
   PTree_p      parents = NULL;
   long         clause_id;
   long         parent_no;
   CompClause_p parent;
   NumTree_p     cell;
   ClauseStats_p stats;

   AcceptInpId(in, "CR");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   if(PDArrayElementP(state->clause_store, clause_id))
   {
      AktTokenError(in, "Target clause of renaming already exists", false);
   }
   /*  printf("CR %ld\n", clause_id); */
   NextToken(in);
   AcceptInpId(in, "RENAME");
   parent_no = parse_parent_list(in, state, &parents);
   if(parent_no != 1)
   {
      AktTokenError(in,
		    "RENAME expression takes exactly one argument",
		    false);
   }
   assert(parents);
   parent = parents->key;
   
   clause = UnpackClause(parent, state->terms);
   clause->ident = clause_id;    
   compact = CompactifyClause(clause);
   compact->g_parents = parents;
   if((cell = NumTreeFind(&(state->watch_clauses), clause_id)))
   {      
      ClauseSetProp(compact, CPIsWatched);
      stats = cell->val1.p_val;
      compact->stats = *stats;
   }
   PDArrayAssignP(state->clause_store, clause_id, compact);   
   InfStateAddTermAssoc(state, compact);
   
   ClauseSetProp(parent, CPDeleteClause);
   InfStateRemoveTermAssoc(state, parent);
   CompClauseRemoveTerms(parent, state->terms);
   assert(verify_clause_term_assoc(state, compact, "CR"));
}

/*-----------------------------------------------------------------------
//
// Function: InfStateCPParse()
//
//   Parse a CP (Clause processing) inference.
//
// Global Variables: -
//
// Side Effects    : Adds some properties to clauses.
//
/----------------------------------------------------------------------*/

void InfStateCPParse(Scanner_p in, InfState_p state)
{
   long clause_id;
   CompClause_p compact;

   AcceptInpId(in, "CP");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   compact = PDArrayElementP(state->clause_store, clause_id);
   if(!compact)
   {
      AktTokenError(in, "Unknown clause selected for processing",
		    false);
   }
   NextToken(in);
   assert(clause_id == compact->ident);
   CheckInpId(in, CP_PROCESS "|" CP_PROOF "|" CP_FINAL);
   if(TestInpId(in, CP_PROCESS))
   {
      ClauseSetProp(compact, CPIsSelected);
   }
   else if(TestInpId(in, CP_PROOF))
   {
      ClauseSetProp(compact, CPIsFinal|CPIsSelected);
   }
   else if(TestInpId(in, CP_FINAL))
   {
      ClauseSetProp(compact, CPIsFinal);
   }
   else
   {
      assert(false);
   }
   NextToken(in);
   assert(verify_clause_term_assoc(state, compact, "CP"));
}

/*-----------------------------------------------------------------------
//
// Function: InfStateCSParse()
//
//   Parse a CS (Clause status change) inference.
//
// Global Variables: -
//
// Side Effects    : May change status of clause.
//
/----------------------------------------------------------------------*/

void InfStateCSParse(Scanner_p in, InfState_p state)
{
   long clause_id;
   CompClause_p compact;

   AcceptInpId(in, "CS");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   compact = PDArrayElementP(state->clause_store, clause_id);
   if(!compact)
   {
      AktTokenError(in, "Unknown clause specified for status change",
		    false);
   }
   assert(verify_clause_term_assoc(state, compact, "CS"));
   /* {
      int i;
      printf("CS %ld", clause_id);
      for(i=0; i<compact->literal_no; i++)
      {
	 printf(" : %ld=%ld ", compact->lit_terms[2*i]->entry_no, 
		compact->lit_terms[(2*i)+1]->entry_no);	  
      }
      printf("\n");
   }*/
   NextToken(in);
   assert(clause_id == compact->ident);
   
   AcceptInpId(in, CS_DELETE);
   ClauseSetProp(compact, CPDeleteClause);
   InfStateRemoveTermAssoc(state, compact);
   CompClauseRemoveTerms(compact, state->terms);
}

/*-----------------------------------------------------------------------
//
// Function: InfStateCMParse()
//
//   Parse a CM (Clause Modify) inference.
//
// Global Variables: -
//
// Side Effects    : Changes term associations
//
/----------------------------------------------------------------------*/

void InfStateCMParse(Scanner_p in, InfState_p state)
{
   Clause_p     clause;
   CompClause_p compact;
   long         clause_id;
   bool         ext_rep_change = false;
   
   AcceptInpId(in, "CM");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   compact = PDArrayElementP(state->clause_store, clause_id);
   if(!compact)
   {
      AktTokenError(in, "Unknown clause specified for modification",
		    false);
   }
   NextToken(in);
   /* printf("CM %ld\n", clause_id); */
   assert(clause_id == compact->ident);
   
   if(!TestInpId(in, CM_TERM_RENAME))
   {
      ext_rep_change = true;
   }
   AcceptInpId(in, CM_SIMPLIFY_REFLECT "|" CM_TERM_RENAME "|" CM_MINIMIZE);
   parse_parent_list(in, state, &(compact->g_parents));
   clause = ClauseParse(in, state->terms);
   InfStateRemoveTermAssoc(state, compact);
   CompClauseRemoveTerms(compact, state->terms);
   CompClauseAddTerms(compact, clause);
   InfStateAddTermAssoc(state, compact);   
   ClauseFree(clause);
   if(state->print_clause_mods && ext_rep_change)
   {
      print_example_conditional(compact, state);
   }
   assert(verify_clause_term_assoc(state, compact, "CM"));
}


/*-----------------------------------------------------------------------
//
// Function: InfStateCEParse()
//
//   Parse a CE (Clause Evaluate) inference.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void InfStateCEParse(Scanner_p in, InfState_p state)
{
   CompClause_p compact;
   long         clause_id;
   
   AcceptInpId(in, "CE");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   compact = PDArrayElementP(state->clause_store, clause_id);
   if(!compact)
   {
      AktTokenError(in, "Unknown clause for evaluation",
		    false);
   }
   NextToken(in);
   assert(clause_id == compact->ident);
   
   if(state->print_clause_evals)
   {
      print_example_conditional(compact, state);
   }
   assert(verify_clause_term_assoc(state, compact, "CE"));
}


/*-----------------------------------------------------------------------
//
// Function: InfStateRWParse()
//
//   Parse a RW (rewrite) inference into state.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void InfStateRWParse(Scanner_p in, InfState_p state)
{
   long clause_id;
   Term_p term, repl, t1, r1;
   CompClause_p compact;
   PStack_p super_stack = PStackAlloc(), tmp_stack;
   PTree_p  affected_clauses = NULL;

   AcceptInpId(in, "RW");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   compact = PDArrayElementP(state->clause_store, clause_id);
   if(!compact)
   {
      AktTokenError(in, "Unknown clause specified as rewrite-rule",
		    false);
   }
   NextToken(in);
   assert(clause_id == compact->ident);
   
   term = TBTermParse(in, state->terms);
   AcceptInpTok(in, LesserSign);
   AcceptInpTokNoSkip(in, EqualSign);
   repl = TBTermParse(in, state->terms);

   if(state->print_clause_mods)
   {
      find_associated_clauses(state, term, &affected_clauses);
   }
   
   tmp_stack = PStackAlloc();
   PStackPushP(tmp_stack, term);
   PStackPushP(tmp_stack, repl);
   parse_term_list(in, state->terms, tmp_stack);
   PStackPushP(super_stack, tmp_stack);

   rek_store_rewrite_parent(state, term, compact);
   change_term_clause_association(state, term, repl);
   while(!TestInpTok(in, Fullstop))
   {
      t1 = TBTermParse(in, state->terms);
      AcceptInpTok(in, LesserSign);
      AcceptInpTokNoSkip(in, EqualSign);
      r1 = TBTermParse(in, state->terms);
      tmp_stack = PStackAlloc();
      PStackPushP(tmp_stack, t1);
      PStackPushP(tmp_stack, r1);
      parse_term_list(in, state->terms, tmp_stack);
      PStackPushP(super_stack, tmp_stack);
      change_term_clause_association(state, t1, r1);
   }
   AcceptInpTok(in, Fullstop);
   TBTermReplaceRepro(state->terms, term, repl, super_stack);
   
   while(!PStackEmpty(super_stack))
   {
      PStackFree(PStackPopP(super_stack));
   }
   PStackFree(super_stack);

   if(state->print_clause_mods)
   {
      /* printf("RW %ld\n", clause_id); */
      print_example_tree(affected_clauses, state);
      /* printf("...done\n"); */
      PTreeFree(affected_clauses);
   }   
}



/*-----------------------------------------------------------------------
//
// Function: InfStateInferenceParse() 
//
//   Parse an arbitrary inference into state.
//
// Global Variables: -
//
// Side Effects    : Reads input, changes state
//
/----------------------------------------------------------------------*/

void InfStateInferenceParse(Scanner_p in, InfState_p state)
{
   if(TestInpId(in, "CC"))
   {
      InfStateCCParse(in, state);
   }
   else if(TestInpId(in, "CP"))
   {
      InfStateCPParse(in, state);
   }
   else if(TestInpId(in, "CS"))
   {
      InfStateCSParse(in, state);
   }
   else if(TestInpId(in, "CM"))
   {
      InfStateCMParse(in, state);
   }
   else if(TestInpId(in, "CR"))
   {
      InfStateCRParse(in, state);
   }
   else if(TestInpId(in, "CE"))
   {
      InfStateCEParse(in, state);
   }
   else if(TestInpId(in, "RW"))
   {
      InfStateRWParse(in, state);
   }
   else
   {
      AcceptInpId(in, "CC|CP|CM|RW|CS|CR|CE");
      assert(false && "Not yet implemented");
   }
}


/*-----------------------------------------------------------------------
//
// Function: InfStateInfListParse()
//
//   Parse a list of inference step descriptions into the
//   state. Return number of steps parsed.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long InfStateInfListParse(Scanner_p in, InfState_p state)
{
   long steps = 0;

   while(!TestInpTok(in,NoToken))
   {
      InfStateInferenceParse(in, state);
      steps++;
   }
   return steps;
}


/*-----------------------------------------------------------------------
//
// Function: InfStatePrintClauses()
//
//   Print the selected clauses and information about them.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void InfStatePrintClauses(FILE* out, InfState_p state,
			  ClauseProperties filter, bool status, bool
			  deps, bool stats)
{
   long i;
   CompClause_p clause;

   for(i=0; i<state->clause_store->size; i++)
   {
      clause = PDArrayElementP(state->clause_store,i);
      if(clause && ClauseQueryProp(clause, filter))
      {
	 fprintf(out, "%ld", i);
	 if(status)
	 {
	    fprintf(out, ":%c,%c",
		    ClauseQueryProp(clause, CPIsSelected)?'S':'N',
		    ClauseQueryProp(clause, CPIsProofClause)?'P':'N');
	 }
	 if(deps)
	 {
	    fprintf(out, ":");
	    print_parent_tree(out, clause->g_parents);
	    print_parent_tree(out, clause->s_parents);	 
	 }
	 if(stats)
	 {
	    fprintf(out, ":");
	    ClauseStatsPrint(out, &(clause->stats));
	 }
	 fprintf(out, "\n");
      }
   }
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/













