/*-----------------------------------------------------------------------

File  : ccl_clauses.c

Author: Stephan Schulz

Contents
 
  Clauses - Infrastructure functions

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Thu Apr 16 19:38:16 MET DST 1998
    New

<2> Wed Oct 14 08:40:21 MET DST 1998
    Extended (ClauseNotGreaterEqual, foundEqLitLater),
         Joachim Steinbach

-----------------------------------------------------------------------*/


#include "ccl_clauses.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

bool   ClausesHaveLocalVariables = true; /* Reassign variable numbers
					    for each clause. This
					    optimizes sharing, but
					    will also loose
					    inter-clause relationships
					    needed for the analysis
					    program */

static long global_clause_counter = 0; /* Counts calls to
					  ClauseAlloc()  and
					  ClauseCopy */

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: foundEqLitLater
//
// Checks whether there is a literal (say L) in lits2 which is
// equivalent wrt. an ordering (given in ocb) to the first literal in
// lits1 (say M). In this case, the property `EPHasEquiv' is set for L
// and M.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static bool foundEqLitLater(OCB_p ocb, Eqn_p lits1, Eqn_p lits2)
{
   while (lits2)
   {
      if ((!EqnHasEquiv(lits2)) &&
	  (LiteralCompare(ocb, lits1, lits2) == to_equal))
      {
	 EqnSetProp(lits1, EPHasEquiv);
	 EqnSetProp(lits2, EPHasEquiv);
	 return true;
      }
      lits2 = lits2->next;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: term_query_var_prop() 
//
//   For all subterms with todo set check all variables for query,
//   return true if one is found, false otherwise.
//

// Global Variables: -
//
// Side Effects    : Changes TPOpFlag
//
/----------------------------------------------------------------------*/

static bool term_query_var_prop(Term_p term, TermProperties query,
			 TermProperties todo)
{
   PStack_p stack = PStackAlloc();
   bool     res = false;
   int      i;

   PStackPushP(stack, term);
   
   while(!PStackEmpty(stack))
   {
      term = PStackPopP(stack);
      if(TermCellIsAnyPropSet(term, todo))
      {
	 TermCellDelProp(term, todo);
	 if(TermIsVar(term))
	 {
	    if(TermCellQueryProp(term, query))
	    {
	       res = true;
	       break;
	    }
	 }
	 else
	 {
	    for(i=0; i<term->arity; i++)
	    {
	       PStackPushP(stack, term->args[i]);
	    }
	 }
      }
   }
   PStackFree(stack);
   return res;
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: ClauseSetTPTPType()
//
//   Set the TPTP type of a clause. 
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void ClauseSetTPTPType(Clause_p clause, ClauseProperties type)
{
   ClauseDelProp(clause,CPTypeMask);
   ClauseSetProp(clause, type);
}


/*-----------------------------------------------------------------------
//
// Function: EmptyClauseAlloc()
//
//   Return a pointer to an empty clause initialized with rational
//   values.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

Clause_p EmptyClauseAlloc()
{
   Clause_p handle = ClauseCellAlloc();

   handle->literals    = NULL;
   handle->neg_lit_no  = 0;
   handle->pos_lit_no  = 0;
   handle->evaluations = NULL;
   handle->set         = NULL;
   handle->properties  = CPIgnoreProps;
   handle->proof_depth = 0;
   handle->proof_size  = 0;
   handle->children    = NULL;
   handle->parent1     = NULL;
   handle->parent2     = NULL;
   handle->pred        = NULL;
   handle->succ        = NULL;

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseAlloc()
//
//   Create a new clause with the literal list list. Does sort literal
//   list by pos/neg-literals for easier comparison, does not use
//   EqnList functions because I'm a stupid performance freak.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Clause_p ClauseAlloc(Eqn_p literals)
{
   Clause_p handle = EmptyClauseAlloc();
   Eqn_p pos_lits=NULL, *pos_append = &pos_lits;
   Eqn_p neg_lits=NULL, *neg_append = &neg_lits;
   Eqn_p next;

   handle->ident = ++global_clause_counter;

   while(literals)
   {      
      next = literals->next;
      if(EqnIsPositive(literals))
      {
	 handle->pos_lit_no++;
	 *pos_append = literals;
	 pos_append = &((*pos_append)->next);
      }
      else
      {
	 handle->neg_lit_no++;
	 *neg_append = literals;
	 neg_append = &((*neg_append)->next);
       }
      literals = next;
   }
   *pos_append = neg_lits;
   *neg_append = NULL;
   handle->literals = pos_lits; 

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseFree()
//
//   Free a clause. Does not take care of parents, children, etc., but
//   releases the memory directly taken by the clause.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseFree(Clause_p junk)
{
   assert(!junk->set);
   EvalListFree(junk->evaluations);   
   EqnListFree(junk->literals);
   PTreeFree(junk->children);
   ClauseCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseIsRangeRestricted()
//
//   Return true if clause is range-restricted, i.e. if all variables
//   occuring in the tail also occur in the head. 
//
// Global Variables: -
//
// Side Effects    : Changes TPOpFlag und TPSpecialFlag in affected
//                   term cells
//
/----------------------------------------------------------------------*/

bool ClauseIsRangeRestricted(Clause_p clause)
{
   Eqn_p    handle;

   if(ClauseIsPositive(clause))
   {
      return true;
   }
   else if(ClauseIsNegative(clause))
   {
      return false;
   }
   for(handle=clause->literals; handle; handle = handle->next)
   {
      if(EqnIsNegative(handle))
      {
	 TermSetProp(handle->lterm, DEREF_NEVER,
		     TPOpFlag|TPSpecialFlag);
	 TermSetProp(handle->rterm, DEREF_NEVER,
		     TPOpFlag|TPSpecialFlag);
      }
   }
   for(handle=clause->literals; handle; handle = handle->next)
   {
      if(EqnIsPositive(handle))
      {
	 TermDelProp(handle->lterm, DEREF_NEVER, TPSpecialFlag);
	 TermDelProp(handle->rterm, DEREF_NEVER, TPSpecialFlag);
      }
   }
   for(handle=clause->literals; handle; handle = handle->next)
   {
      if(EqnIsNegative(handle))
      {
	 if(term_query_var_prop(handle->lterm, TPOpFlag, TPSpecialFlag))
	 {
	    return false;
	 }
	 if(term_query_var_prop(handle->rterm, TPOpFlag, TPSpecialFlag))
	 {
	    return false;
	 }	 
      }
   }
   return true;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseCopy()
//
//   Create a semantically equivalent clause and return a pointer to
//   it. Does not copy parents, children, etc. Terms in the original
//   clause are interpreted as instantiated, and are created in or
//   retrived from the new bank. Evaluations are not copied. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Clause_p ClauseCopy(Clause_p clause, TB_p bank)
{
   Clause_p handle = ClauseCellAlloc();

   handle->literals    = EqnListCopy(clause->literals, bank);
   handle->ident       = ++global_clause_counter;
   handle->neg_lit_no  = clause->neg_lit_no;
   handle->pos_lit_no  = clause->pos_lit_no;
   handle->evaluations = NULL;
   handle->set         = NULL;
   handle->properties  = clause->properties;
   handle->proof_depth = clause->proof_depth;
   handle->proof_size  = clause->proof_size;
   handle->children    = NULL;
   handle->parent1     = NULL;
   handle->parent2     = NULL;
   handle->pred        = NULL;
   handle->succ        = NULL;

   return handle;
}



/*-----------------------------------------------------------------------
//
// Function: ClausePrintList()
//
//   Print a clause as a declarative list of literals.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintList(FILE* out, Clause_p clause, bool fullterms)
{
   EqnListPrint(out, clause->literals, "; ", normal, fullterms);
   fprintf(out, " <-.");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintAxiom()
//
//   Print a clause as a declarative axiom in normal form, i.e. print
//   positive literals first, then <-, then negative literals.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintAxiom(FILE* out, Clause_p clause, bool fullterms)
{
   Eqn_p handle;
   int   i;

   i = 0;
   for(handle = clause->literals; handle; handle = handle->next)
   {
      if(EqnIsPositive(handle))
      {
	 EqnPrint(out, handle, normal, fullterms);
	 i++;
	 if(i<clause->pos_lit_no)
	 {
	    fprintf(out, "; ");
	 }
      }
   }
   fprintf(out, " <- ");
   
   i = 0;
   for(handle = clause->literals; handle; handle = handle->next)
   {
      if(!EqnIsPositive(handle))
      {
	 EqnPrint(out, handle, special, fullterms);
	 i++;
	 if(i<clause->neg_lit_no)
	 {
	    fprintf(out, ", ");
	 }
      }
   }
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintRule()
//
//   Print a clause as a rule, with the head literal as the conclusion
//   and the remaining literals as preconditions.  If a
//   clause has a single literal only, print it as a fact.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintRule(FILE* out, Clause_p clause, bool fullterms)
{
   if(clause->literals)
   {
      EqnPrint(out, clause->literals, normal, fullterms);
      if(clause->literals->next)
      {
	 fprintf(out, " <- ");
	 EqnListPrint(out, clause->literals->next, ", ", special,
		      fullterms);
      }
   }
   else
   {
      fprintf(out, " <- ");
   }
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintGoal()
//
//   Print a clause as a goal, i.e. put all literals behind <-.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintGoal(FILE* out, Clause_p clause, bool fullterms)
{
   fprintf(out, "<- ");
   EqnListPrint(out, clause->literals, ", ", special, fullterms);
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintQuery()
//
//   Print a clause as a procedural query, i.e. put all literals
//   behind ?-.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintQuery(FILE* out, Clause_p clause, bool fullterms)
{
   assert(ClauseLiteralNumber(clause));

   fprintf(out, "?- ");
   EqnListPrint(out, clause->literals, ", ", special, fullterms);
   fprintf(out, ".");
}

/*-----------------------------------------------------------------------
//
// Function: ClausePrintTPTPFormat()
//
//   Print a clause in TPTP format.  
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintTPTPFormat(FILE* out, Clause_p clause)
{
   char* typename;

   switch(ClauseQueryTPTPType(clause))
   {
   case CPTypeAxiom:
      typename = "axiom";
      break;
   case CPTypeHypothesis:
      typename = "hypothesis";
      break;      
   case CPTypeConjecture:
      typename = "conjecture";
      break;
   default:
	 typename = "unknown";
	 break;
   }   

   fprintf(out, "input_clause(c%ld,%s,[",
	   clause->ident, typename);
   EqnListPrint(out, clause->literals, ",", false, true);
   fprintf(out, "]).\n");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrint()
//
//   Print a clause in the most canonical representation.
//
// Global Variables: TPTPFormatPrint
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrint(FILE* out, Clause_p clause, bool fullterms)
{
   DEBUG(65535, fprintf(out, "/*(%3ld)*/ ", clause->ident););
   /* fprintf(out, "(%ld, %ld)", clause->proof_depth,
      clause->proof_size); */
   if(TPTPFormatPrint)
   {
      ClausePrintTPTPFormat(out, clause);
   }
   else
   {
      /* 
      if(clause->pos_lit_no || !clause->neg_lit_no)
      {
	 ClausePrintAxiom(out, clause, fullterms);
      }
      else
      {
	 ClausePrintQuery(out, clause, fullterms);
      }*/
      ClausePrintAxiom(out, clause, fullterms);      
   }
   DEBUG(64, EvalListPrintComment(out, clause->evaluations););
   DEBUG(64, printf("Address: %d", (int)clause););
}


/*-----------------------------------------------------------------------
//
// Function: ClauseStartsMaybe()
//
//   Return true if a clause possibly starts on the current position
//   in the input, i.e. if TermStartToken, TildeSign, ?-, or <- are
//   present on the input stream.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseStartsMaybe(Scanner_p in)
{
   if(TestInpTok(in, TermStartToken|TildeSign))
   {
      return true;
   }
   if(TestInpTok(in, LesserSign|QuestionMark))
   {
      if(!LookToken(in,1)->skipped && TestTok(LookToken(in, 1), Hyphen))
      {
	 return true;
      }
   }
   return false;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseParse()
//
//   Parse a clause.
//
// Global Variables: -
//
// Side Effects    : Input, memory operations
//
/----------------------------------------------------------------------*/

Clause_p ClauseParse(Scanner_p in, TB_p bank)
{
   Eqn_p    concl, precond;
   bool     procedural = false;
   ClauseProperties type = CPTypeAxiom;
   Clause_p handle;

   if(ClausesHaveLocalVariables)
   {
      VarBankClearExtNames(bank->vars);
   }   
   if(TPTPFormatParse)
   {      
      AcceptInpId(in, "input_clause");
      AcceptInpTok(in, OpenBracket);
      AcceptInpTok(in, Name);
      AcceptInpTok(in, Comma);
      CheckInpId(in, "axiom|hypothesis|conjecture");
      if(TestInpId(in, "axiom"))
      {
	 type = CPTypeAxiom;
      }
      else if(TestInpId(in, "conjecture"))
      {
	 type = CPTypeConjecture;
      }
      else
      {
	 type = CPTypeHypothesis;
      }
      AcceptInpTok(in, Ident);
      AcceptInpTok(in, Comma);
      AcceptInpTok(in, OpenSquare);
      concl = EqnListParse(in, bank, Comma);
      AcceptInpTok(in, CloseSquare);
      AcceptInpTok(in, CloseBracket);   
   }
   else
   {
      concl = EqnListParse(in, bank, Semicolon);
      if(TestInpTok(in, Colon))
      {
	 if(EqnListLength(concl)>1)
	 {
	 AktTokenError(in, 
		       "Procedural rule cannot have more than one "
		       "head literal",
		       false);
	 }
	 procedural = true;
      }
      else if(TestInpTok(in, QuestionMark))
      {
	 if(EqnListLength(concl)>0)
	 {
	    AktTokenError(in, 
			  "Query should consist only of tail literals",
			  false);
	 }
	 procedural = true;
      }
      if(TestInpTok(in, Fullstop))
      {
	 if(EqnListLength(concl)>1)
	 {
	    AktTokenError(in, 
			  "Procedural fact cannot have more than one "
			  "literal",
			  false);
	 }
	 procedural = true;
      }
      else
      {
	 if(!EqnListLength(concl))
	 {
	    type = CPTypeConjecture;
	 }
	 AcceptInpTok(in, LesserSign | Colon | QuestionMark);
	 AcceptInpTokNoSkip(in, Hyphen);
	 
	 precond = EqnListParse(in, bank, Comma);
	 
	 if(procedural && EqnListLength(precond)==0)
	 {
	    AktTokenError(in, 
			  "Procedural rule or query needs at least one "
			  "tail literal (Hey! I did not make this "
			  " syntax! -StS)", 
			  false);
	 }
	 EqnListNegateEqns(precond);
	 EqnListAppend(&concl, precond);
      }
   }
   AcceptInpTok(in, Fullstop);
   handle = ClauseAlloc(concl);
   ClauseSetTPTPType(handle, type);
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseMarkMaximalTerms()
//
//   Orient literals, mark maximal literals.
//
// Global Variables: -
//
// Side Effects    : Changes clause structure, orients equations
//
/----------------------------------------------------------------------*/

void ClauseMarkMaximalTerms(OCB_p ocb, Clause_p clause)
{
   EqnListOrient(ocb, clause->literals);
   EqnListMaximalLiterals(ocb, clause->literals);
   ClauseSetProp(clause, CPIsOriented);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseMarkRestrictedTerms()
//
//   Mark terms on which rewriting is restricted.
//
// Global Variables: -
//
// Side Effects    : Changes term bank...
//
/----------------------------------------------------------------------*/

void ClauseMarkRestrictedTerms(Clause_p clause)
{
   EqnListMarkRestrictedTerms(clause->literals);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseParentsAreSubset()
//
//   Return true if parents of clause1 are a subset of clause2. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseParentsAreSubset(Clause_p clause1, Clause_p clause2)
{
   bool sub1 = false, sub2 = false;

   if(clause1->parent1)
   {
      if((clause1->parent1==clause2->parent1)||
	 (clause1->parent1==clause2->parent1))
      {
	 sub1=true;
      }
   }
   else
   {
      sub1=true;
   }
   if(clause1->parent2)
   {
      if((clause1->parent2==clause2->parent1)||
	 (clause1->parent2==clause2->parent1))
      {
	 sub2=true;
      }
   }
   else
   {
      sub2=true;
   }
   return sub1&&sub2;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseDetachParents()
//
//   Discouple a clause from it's parents, i.e. make the parents
//   forget the clause and vice versa.
//
// Global Variables: -
//
// Side Effects    : Changes parens children-tree.
//
/----------------------------------------------------------------------*/

void ClauseDetachParents(Clause_p clause)
{
   bool result;

   if(clause->parent1)
   {
      result = PTreeDeleteEntry(&(clause->parent1->children), clause);
      assert(result);
      clause->parent1 = NULL;
   }
   if(clause->parent2)
   {
      result = PTreeDeleteEntry(&(clause->parent2->children), clause);
      assert(result);
      clause->parent2 = NULL;
   }   
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRegisterChild()
//
//   Make child a child of clause. Clause already should be a parent
//   to child!
//
// Global Variables: -
//
// Side Effects    : Changes clause->children
//
/----------------------------------------------------------------------*/

void ClauseRegisterChild(Clause_p clause, Clause_p child)
{
   assert(child->parent1==clause || child->parent2==clause);
   
   PTreeStore(&(clause->children), child);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseAddEvalCell()
//
//   Add an evaluation cell (as the first evaluation) to a clause. 
//
// Global Variables: -
//
// Side Effects    : Changes eval list, makes evaluation->object point
//                   to clause.
//
/----------------------------------------------------------------------*/

void ClauseAddEvalCell(Clause_p clause, Eval_p evaluation)
{
   evaluation->object  = clause;
   evaluation->next    = clause->evaluations;
   clause->evaluations = evaluation;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveEvaluations()
//
//   Remove the evaluations from a clause and free the EvalCells.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseRemoveEvaluations(Clause_p clause)
{
   EvalListFree(clause->evaluations);
   clause->evaluations = NULL;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseWeight()
//
//   Compute the weight of a clause by counting function symbols and
//   variables and applying various modifiers.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

double ClauseWeight(Clause_p clause, double max_term_multiplier,
		    double max_literal_multiplier, double
		    pos_multiplier, long vweight, long fweight, bool
		    count_eq_encoding)
{
   Eqn_p  handle;
   double res = 0;

   for(handle = clause->literals; handle; handle = handle->next)
   {
      res += LiteralWeight(handle, max_term_multiplier,
			   max_literal_multiplier, pos_multiplier,
			   vweight, fweight, count_eq_encoding);
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseStandardWeight()
//
//   Compute the standard weight of a clause (Vars = 1, Funs = 2,
//   everything counts equally.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

double ClauseStandardWeight(Clause_p clause)
{
   Eqn_p  handle;
   double res = 0;

   for(handle = clause->literals; handle; handle = handle->next)
   {
      res += EqnStandardWeight(handle);
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseOrientWeight()
//
//   Compute the weight of a clause by counting function symbols and
//   variables and applying various other modifiers ;-).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

double ClauseOrientWeight(Clause_p clause, double
			  unorientable_literal_multiplier,
			  double max_literal_multiplier, double
			  pos_multiplier, long vweight, long fweight, bool
			  count_eq_encoding)
{
   Eqn_p  handle;
   double res = 0 ,tmp;
   
   for(handle = clause->literals; handle; handle = handle->next)
   {
      tmp = LiteralWeight(handle, 1, max_literal_multiplier,
			  pos_multiplier, vweight, fweight,
			  count_eq_encoding);
      if(!EqnIsOriented(handle))
      {
	 tmp = tmp*unorientable_literal_multiplier;
      }
      res += tmp;
   }
   return res;


}


/*-----------------------------------------------------------------------
//
// Function: ClauseNotGreaterEqual
//
//   Checks whether the multiset of literals of a clause is neither
//   greater than (wrt. the multiset extension of a specified
//   ordering) nor equal to the multiset of literals of another
//   clause.
//
//   not(A >=>= B)  iff  - not(A = B) and
//                       - for each b in B\A there does not exist an a
//                         in A\B such that a > b
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseNotGreaterEqual(OCB_p ocb,
			   Clause_p clause1, Clause_p clause2)
{
   Eqn_p cl1lits = clause1->literals;
   Eqn_p cl2lits = clause2->literals, cl2litseqs;
   bool foundeq, foundeqlater, foundgtr;

   EqnListDelProp(cl1lits, EPHasEquiv | EPDominates);

   while (cl2lits)
   {
      EqnDelProp(cl2lits, EPHasEquiv | EPIsDominated);
      foundeq = foundgtr = false;

      while (cl1lits && !EqnHasEquiv(cl2lits))
      {
	 if (!EqnHasEquiv(cl1lits))
	 {
	    switch (LiteralCompare(ocb, cl1lits, cl2lits))
	    {
	    case to_greater:
		  foundeqlater = false;
		  if (!EqnDominates(cl1lits))
		  {
		     cl2litseqs = cl2lits->next;
		     foundeqlater =
			foundEqLitLater(ocb, cl1lits, cl2litseqs);
		  }
		  if (!foundeqlater)
		  {
		     EqnSetProp(cl1lits, EPDominates);
		     EqnSetProp(cl2lits, EPIsDominated);
		     foundgtr = true;		     
		  }
		  break;
	    case to_equal:
		  EqnSetProp(cl1lits, EPHasEquiv);
		  EqnSetProp(cl2lits, EPHasEquiv);
		  foundeq = true;
		  break;
	    case to_lesser:
	    case to_uncomparable:
		  break;
	    default:
		  assert(false);
		  break;
	    }
	 }
	 cl1lits = cl1lits->next;
      }

      /* early abortion if there is neither a greater nor an equal
	 element in clause1 */
      if ((!foundeq) && (!foundgtr))
      {
	 /* printf("True\n"); */
	 return true;
      }
      cl2lits = cl2lits->next;
   }
   /* printf("False\n"); */
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseCompareFun()
//
//   Compare two clauses, induce a total ordering on all clauses. If
//   ClauseCompareFun(clause1, clause2) == 0, then clause1==clause2
//   modulo symmetry of = and AC of the disjunction (In the current
//   implementation, only symmetry is taken into account.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

int ClauseCompareFun(Clause_p clause1, Clause_p clause2)
{
   int res;
   Eqn_p handle1, handle2;

   if((res = clause2->pos_lit_no - clause1->pos_lit_no))
   {
      return res;
   }
   if((res = clause2->neg_lit_no - clause1->neg_lit_no))
   {
      return res;
   }
   handle1 = clause1->literals;
   handle2 = clause2->literals;
   while(handle1)
   {
      assert(handle2);
      res = LiteralCompareFun(handle1, handle2);
      if(res)
      {
	 break;
      }
      handle1 = handle1->next;
      handle2 = handle2->next;
   }
   return res;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/
