/*-----------------------------------------------------------------------

File  : ccl_pdtrees.h

Author: Stephan Schulz

Contents
 
  Perfect discrimination trees for optimized rewriting.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Mon Jun 22 17:04:32 MET DST 1998
    New

-----------------------------------------------------------------------*/

#ifndef CCL_PDTREES

#define CCL_PDTREES

#include <clb_pdarrays.h>
#include <ccl_clausepos.h>

/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/


/* A cell in the perfect discrimination tree... */

typedef struct pdtreecell
{
   struct pdtreecell  **f_alternatives; /* function alternatives */
   PDArray_p          v_alternatives;   /* Variables */
   FunCode            max_var;          /* Largest variable index
					   (negated f_code) occuring
					   at this node */
   Term_p             variable;         /* If variable, which one? */
   long               min_lside_weight; /* Smallers left side in
					   clauses at or beyond this
					   node */ 
   SysDate            max_demod_date;   /* Date of the youngest clause
					   at or beyond this node */
   struct pdtreecell  *parent;          /* Where did we come from? */
   FunCode            sig_size;         /* Number of function symbol
					   alternatives */
   long               ref_count;        /* How many terms share this
					   node? */
   PTree_p            demod_terms;      /* Denoted by ClausePosCells */
   FunCode            trav_state;       /* For traversing during
					   matching */
   PStackPointer      var_state;        /* For storing information
					   about variables bindings */
}PDTreeCell, *PDTree_p;

/* A PDSearch state, just to keep all things in one place */

typedef struct pdsearchcell
{
   long     term_weight;   /* Weight of the whole term to match */
   long     weight_offset; /* Parts of the term already "used up"
			      don't count anymore */
   SysDate  term_nf_date;  /* NF-Date of the whole term with respect
			      to the clauses in the current tree */
   PDTree_p trav_pointer;
   PStack_p term_state;
   PStack_p term_proc;
   PStack_p demod_store;
}PDSearchCell, *PDSearch_p;


/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/

extern bool PDTreePreferGeneral;

#define  PDTREE_IGNORE_TERMWEIGHTS LONG_MAX
#define  PDTREE_IGNORE_NF_DATE     SysDateCreationTime()

#define  PDTreeCellAlloc()    (PDTreeCell*)SizeMalloc(sizeof(PDTreeCell))
#define  PDTreeCellFree(junk) SizeFree(junk, sizeof(PDTreeCell))
PDTree_p PDTreeAlloc(FunCode sig_size);
void     PDTreeDefaultCellFree(PDTree_p root);
void     PDTreeFree(PDTree_p root);

#define    PDSearchCellAlloc() (PDSearchCell*)SizeMalloc(sizeof(PDSearchCell))
#define    PDSearchCellFree(junk) SizeFree(junk, sizeof(PDSearchCell))
PDSearch_p PDSearchAlloc(PDTree_p index, Term_p term, SysDate nf_date);
void       PDSearchFree(PDSearch_p junk);

static __inline__ PStack_p TermLRTraverseInit(Term_p term);
static __inline__ Term_p   TermLRTraverseNext(PStack_p stack);
static __inline__ Term_p   TermLRTraversePrev(PStack_p stack, Term_p term);
#define  TermLRTraverseExit(stack) PStackFree(stack)

void     PDTreeInsert(PDTree_p root, ClausePos_p demod_side);
void     PDTreeDelete(PDTree_p root, Term_p term, Clause_p clause);

PDTree_p PDTreeFindNextIndexedTerm(PDSearch_p state, Subst_p subst);

ClausePos_p PDTreeFindNextDemodulator(PDSearch_p state, Subst_p
				      subst);

PDTree_p PDTreeFindTerm(PDTree_p index, Term_p term);
void PDTreeStoreClauses(PTree_p *tree, PDTree_p index, Term_p term);

bool ClauseHasIndexedTerm(Clause_p clause);
long ClauseRemoveFromIndex(Clause_p clause, PDTree_p index);

void PDTreeDebugPrintNode(FILE* out, PDTree_p node);
void PDTreeDebugPrint(FILE* out, PDTree_p node);

/*---------------------------------------------------------------------*/
/*                        Inline-Functions                             */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: TermLRTraverseInit()
//
//   Initialize a stack for left-right term traversal (i.e. traversal
//   in writing order) and return it.
//
// Global Variables: -
//
// Side Effects    : Memory management
//
/----------------------------------------------------------------------*/

static __inline__ PStack_p TermLRTraverseInit(Term_p term)
{
   PStack_p stack = PStackAlloc();
   
   PStackPushP(stack, term);
   
   return stack;
}


/*-----------------------------------------------------------------------
//
// Function: TermLRTraverseNext()
//
//   Return the next term node in LR-ordering and update the
//   stack. Return NULL if term traveral is complete.
//
// Global Variables: -
//
// Side Effects    : Stack changes
//
/----------------------------------------------------------------------*/

static __inline__ Term_p TermLRTraverseNext(PStack_p stack)
{
   int    i;
   Term_p handle;
   
   if(PStackEmpty(stack))
   {
      return NULL;
   }
   handle = PStackPopP(stack);
   for(i=handle->arity-1; i>=0; i--)
   {
      PStackPushP(stack, handle->args[i]);
   }
   
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: TermLRTraversePrev()
//
//   Undo a TermLRTraverseNext() operation by replacing terms args on
//   the stack with term.
//
// Global Variables: -
//
// Side Effects    : Stack changes
//
/----------------------------------------------------------------------*/

static __inline__ Term_p TermLRTraversePrev(PStack_p stack, Term_p term)
{
   Term_p tmp;
   int    i;

   for(i=0; i<term->arity; i++)
   {
      tmp = PStackPopP(stack);
      assert(tmp == term->args[i]);
   }
   PStackPushP(stack, term);

   return term;
}

#endif 

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





