/*-----------------------------------------------------------------------

File  : che_litselection.c

Author: Stephan Schulz

Contents

  Functions for selection certain literals (and hence superposition
  strategies).


  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Fri May 21 22:16:27 GMT 1999
    New

-----------------------------------------------------------------------*/

#include "che_litselection.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

char* LiteralSelectionFunNames[]=
{ 
   "NoSelection",
   "NoGeneration",
   "SelectNegativeLiterals",
   "PSelectNegativeLiterals",
   "SelectPureVarNegLiterals",
   "PSelectPureVarNegLiterals",
   "SelectLargestNegLit",
   "PSelectLargestNegLit",
   "SelectSmallestNegLit",
   "PSelectSmallestNegLit",
   "SelectDiffNegLit",
   "PSelectDiffNegLit",
   "SelectGroundNegLit",
   "PSelectGroundNegLit",
   "SelectOptimalLit",
   "PSelectOptimalLit",
   "SelectCondOptimalLit",
   "PSelectCondOptimalLit",
   "SelectAllCondOptimalLit",
   "PSelectAllCondOptimalLit",
   "SelectOptimalRestrDepth2",
   "PSelectOptimalRestrDepth2",
   "SelectOptimalRestrPDepth2",
   "PSelectOptimalRestrPDepth2",
   "SelectOptimalRestrNDepth2",
   "PSelectOptimalRestrNDepth2",
   "SelectNonRROptimalLit",
   "PSelectNonRROptimalLit",
   NULL
};

static LiteralSelectionFun litsel_fun_array[]=
{
   SelectNoLiterals,
   SelectNoGeneration,
   SelectNegativeLiterals,
   SelectPosNegativeLiterals,
   SelectFirstVariableLiteral,
   SelectPosFirstVariableLiteral,
   SelectLargestNegativeLiteral,
   SelectPosLargestNegativeLiteral,
   SelectSmallestNegativeLiteral,
   SelectPosSmallestNegativeLiteral,
   SelectDiffNegativeLiteral,
   SelectPosDiffNegativeLiteral,
   SelectGroundNegativeLiteral,
   SelectPosGroundNegativeLiteral,
   SelectOptimalLiteral,
   SelectPosOptimalLiteral,          
   SelectCondOptimalLiteral,
   SelectPosCondOptimalLiteral,
   SelectAllCondOptimalLiteral,
   SelectPosAllCondOptimalLiteral,
   SelectDepth2OptimalLiteral,
   SelectPosDepth2OptimalLiteral,
   SelectPDepth2OptimalLiteral,
   SelectPosPDepth2OptimalLiteral,
   SelectNDepth2OptimalLiteral,
   SelectPosNDepth2OptimalLiteral,
   SelectNonRROptimalLiteral,
   SelectPosNonRROptimalLiteral,
   NULL
};


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: GetLitSelFun()
//
//   Given an external name, return a literal selection function or
//   NULL if the name does not match any known function.
//
// Global Variables: LiteralSelectionFunNames, litsel_fun_array
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

LiteralSelectionFun GetLitSelFun(char* name)
{
   int index;

   assert(name);

   index = StringIndex(name, LiteralSelectionFunNames);
   
   if(index>=0)
   {
      return litsel_fun_array[index];
   }
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function: GetLitSelName()
//
//   Given a LiteralSelectionFun, return the corresponding name.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

char* GetLitSelName(LiteralSelectionFun fun)
{
   int  i;
   char *res = NULL;

   for(i=0; litsel_fun_array[i]; i++)
   {
      if(litsel_fun_array[i] == fun)
      {
	 res = LiteralSelectionFunNames[i];
	 assert(res);
	 break;
      }
   }
   assert(res);
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: SelectNoLiterals()
//
//   Unselect all literals.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectNoLiterals(Clause_p clause)
{
   EqnListDelProp(clause->literals, EPIsSelected);
   return;
}

/*-----------------------------------------------------------------------
//
// Function: SelectNoGeneration()
//
//   Do nothing with a clause.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectNoGeneration(Clause_p clause)
{
   assert(clause);

   return;
}


/*-----------------------------------------------------------------------
//
// Function: SelectNegativeLiterals()
//
//   If the clause has negative literals, mark them as selected.
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectNegativeLiterals(Clause_p clause)
{
   assert(clause);

   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;

      while(handle)
      {
	 if(EqnIsPositive(handle))
	 {
	    EqnDelProp(handle, EPIsSelected);
	 }
	 else
	 {
	    EqnSetProp(handle, EPIsSelected);
	 }
	 handle = handle->next;
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPosNegativeLiterals()
//
//   If the clause has negative literals, mark all literals as
//   selected.
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectPosNegativeLiterals(Clause_p clause)
{
   assert(clause);

   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;

      while(handle)
      {
	 EqnSetProp(handle, EPIsSelected);
	 handle = handle->next;
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectFirstVariableLiteral()
//
//   Select first literal of the form X!=Y.
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectFirstVariableLiteral(Clause_p clause)
{
   assert(clause);

   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      
      while(handle)
      {
	 if(EqnIsNegative(handle) && TermIsVar(handle->lterm) &&
	    TermIsVar(handle->rterm))
	 {
	    EqnSetProp(handle, EPIsSelected);
	 }
	 else
	 {
	    EqnDelProp(handle, EPIsSelected);
	 }
	 handle = handle->next;
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPosFirstVariableLiteral()
//
//   If a literal of the form X!=Y exist, select it and all positive
//   literals. Otherwise unselect all literals.
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectPosFirstVariableLiteral(Clause_p clause)
{
   assert(clause);

   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool  found = false;
      
      while(handle)
      {
	 if(EqnIsPositive(handle)||
	    (TermIsVar(handle->lterm) && 
	     TermIsVar(handle->rterm)))
	 {
	    EqnSetProp(handle, EPIsSelected);
	    found = found || EqnIsNegative(handle);
	 }
	 else
	 {
	    EqnDelProp(handle, EPIsSelected);
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectLargestNegativeLiteral()
//
//   Select the largest of the clauses negative literals.
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectLargestNegativeLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals, selected = NULL;
      long  select_weight = 0;
      
      while(handle)
      {
	 EqnDelProp(handle, EPIsSelected);
	 if(EqnIsNegative(handle))
	 {
	    if(EqnStandardWeight(handle) > select_weight)
	    {
	       select_weight = EqnStandardWeight(handle);
	       selected = handle;
	    }
	 }
	 handle = handle->next;
      }
      assert(selected);
      EqnSetProp(selected, EPIsSelected);
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}

/*-----------------------------------------------------------------------
//
// Function: SelectPosLargestNegativeLiteral()
//
//   If clause has negative literals, select the largest of the
//   clauses negative literals and positive literals.
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectPosLargestNegativeLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals, selected = NULL;
      long  select_weight = 0;
      
      while(handle)
      {
	 if(EqnIsPositive(handle))
	 {
	    EqnSetProp(handle, EPIsSelected);
	 }
	 else
	 {
	    EqnDelProp(handle, EPIsSelected);
	    if(EqnStandardWeight(handle) > select_weight)
	    {
	       select_weight = EqnStandardWeight(handle);
	       selected = handle;
	    }
	 }
	 handle = handle->next;
      }
      assert(selected);
      EqnSetProp(selected, EPIsSelected);
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}



/*-----------------------------------------------------------------------
//
// Function: SelectSmallestNegativeLiteral()
//
//   Select the smallest of the clauses negative literals.
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectSmallestNegativeLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals, selected = NULL;
      long  select_weight = LONG_MAX;
      
      while(handle)
      {
	 EqnDelProp(handle, EPIsSelected);
	 if(EqnIsNegative(handle))
	 {
	    if(EqnStandardWeight(handle) < select_weight)
	    {
	       select_weight = EqnStandardWeight(handle);
	       selected = handle;
	    }
	 }
	 handle = handle->next;
      }
      assert(selected);
      EqnSetProp(selected, EPIsSelected);
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}

/*-----------------------------------------------------------------------
//
// Function: SelectPosSmallestNegativeLiteral()
//
//   If clause has negative literals, select the smallest of the
//   clauses negative literals and positive literals.
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectPosSmallestNegativeLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals, selected = NULL;
      long  select_weight = LONG_MAX;
      
      while(handle)
      {
	 if(EqnIsPositive(handle))
	 {
	    EqnSetProp(handle, EPIsSelected);
	 }
	 else
	 {
	    EqnDelProp(handle, EPIsSelected);
	    if(EqnStandardWeight(handle) < select_weight)
	    {
	       select_weight = EqnStandardWeight(handle);
	       selected = handle;
	    }
	 }
	 handle = handle->next;
      }
      assert(selected);
      EqnSetProp(selected, EPIsSelected);
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}

#define lit_sel_diff_weight(handle) \
        ((100*EqnStandardDiff(handle))+EqnStandardWeight(handle))

/*-----------------------------------------------------------------------
//
// Function: SelectDiffNegativeLiteral()
//
//   Select the most unbalanced of the clauses negative literals.
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectDiffNegativeLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals, selected = NULL;
      long  select_weight = -1;
      
      while(handle)
      {
	 EqnDelProp(handle, EPIsSelected);
	 if(EqnIsNegative(handle))
	 {
	    if(lit_sel_diff_weight(handle) > select_weight)
	    {
	       select_weight = lit_sel_diff_weight(handle);
	       selected = handle;
	    }
	 }
	 handle = handle->next;
      }
      assert(selected);
      EqnSetProp(selected, EPIsSelected);
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPosDiffNegativeLiteral()
//
//   If clause has negative literals, select the most unbalanced one
//   of the clauses negative literals and all positive literals.
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectPosDiffNegativeLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals, selected = NULL;
      long  select_weight = -1;
      
      while(handle)
      {
	 if(EqnIsPositive(handle))
	 {
	    EqnSetProp(handle, EPIsSelected);
	 }
	 else
	 {
	    EqnDelProp(handle, EPIsSelected);
	    if(lit_sel_diff_weight(handle) > select_weight)
	    {
	       select_weight = lit_sel_diff_weight(handle);
	       selected = handle;
	    }
	 }
	 handle = handle->next;
      }
      assert(selected);
      EqnSetProp(selected, EPIsSelected);
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectGroundNegativeLiteral()
//
//   If there are negative ground literals, select the one with maximal
//   lit_sel_diff_weight. 
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectGroundNegativeLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals, selected = NULL;
      long  select_weight = -1;
      
      while(handle)
      {
	 EqnDelProp(handle, EPIsSelected);
	 if(EqnIsNegative(handle) && EqnIsGround(handle))
	 {
	    if(lit_sel_diff_weight(handle) > select_weight)
	    {
	       select_weight = lit_sel_diff_weight(handle);
	       selected = handle;
	    }
	 }
	 handle = handle->next;
      }
      if(selected)
      {
	 EqnSetProp(selected, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPosGroundNegativeLiteral()
//
//   If there are negative ground literals, select the one with maximal
//   lit_sel_diff_weight and select all positive literals. 
//
// Global Variables: -
//
// Side Effects    : Changes property in literals
//
/----------------------------------------------------------------------*/

void SelectPosGroundNegativeLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals, selected = NULL;
      long  select_weight = -1;
      
      while(handle)
      {
	 if(EqnIsNegative(handle))
	 {
	    EqnDelProp(handle, EPIsSelected);
	    if(EqnIsGround(handle))
	    {
	       if(lit_sel_diff_weight(handle) > select_weight)
	       {
		  select_weight = lit_sel_diff_weight(handle);
		  selected = handle;
	       }
	    }
	 }
	 else
	 {
	    assert(EqnIsPositive(handle));
	    EqnSetProp(handle, EPIsSelected);
	 }
	 handle = handle->next;
      }
      if(selected)
      {
	 EqnSetProp(selected, EPIsSelected);	 
      }
      else
      {
	 EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectOptimalLiteral()
//
//   (Hah! Believe it at your peril ;-). If there is a ground negative
//   literal, select it, otherwise select the negative literal with the
//   largest size difference.
//
// Global Variables: -
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void SelectOptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals, selected = NULL;
      long  select_weight = -1;
      
      while(handle)
      {
	 EqnDelProp(handle, EPIsSelected);
	 if(EqnIsNegative(handle) && EqnIsGround(handle))
	 {
	    if(lit_sel_diff_weight(handle) > select_weight)
	    {
	       select_weight = lit_sel_diff_weight(handle);
	       selected = handle;
	    }
	 }
	 handle = handle->next;
      }
      if(selected)
      {
	 EqnSetProp(selected, EPIsSelected);
      }
      else
      {
	 SelectDiffNegativeLiteral(clause);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}



/*-----------------------------------------------------------------------
//
// Function: SelectPosOptimalLiteral()
//
//   (Hah! Believe it at your peril ;-). If there is a ground negative
//   literal, select it, otherwise select the negative literal with the
//   largest size difference and all positive literals.
//
// Global Variables: -
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void SelectPosOptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals, selected = NULL;
      long  select_weight = -1;
      
      while(handle)
      {
	 EqnDelProp(handle, EPIsSelected);
	 if(EqnIsNegative(handle) && EqnIsGround(handle))
	 {
	    if(lit_sel_diff_weight(handle) > select_weight)
	    {
	       select_weight = lit_sel_diff_weight(handle);
	       selected = handle;
	    }
	 }
	 handle = handle->next;
      }
      if(selected)
      {
	 EqnSetProp(selected, EPIsSelected);
      }
      else
      {
	 SelectPosDiffNegativeLiteral(clause);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}

/*-----------------------------------------------------------------------
//
// Function: SelectCondOptimalLiteral()
//
//   As above, but if the clause has a positive literal that is very
//   uninstantiated, select no literal at all.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

#define VAR_FACTOR 3

void SelectCondOptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool found = false;
      long weight, sweight;

      while(handle)
      {
	 if(EqnIsPositive(handle))
	 {
	    weight = TermWeight(handle->lterm, 0,VAR_FACTOR);
	    sweight = TermStandardWeight(handle->lterm);
	    
	    if(EqnIsEquLit(handle))
	    {
	       weight += TermWeight(handle->rterm, 0,VAR_FACTOR);
	       sweight += TermStandardWeight(handle->rterm);
	    }
	    if(sweight <= weight)
	    {
	       found = true;
	       break;
	    }
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 SelectOptimalLiteral(clause);
      }
      else
      {
      EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPosCondOptimalLiteral()
//
//   As above, but select positive literals as well.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectPosCondOptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool found = false;
      long weight, sweight;

      while(handle)
      {
	 if(EqnIsPositive(handle))
	 {
	    weight = TermWeight(handle->lterm, 0,VAR_FACTOR);
	    sweight = TermStandardWeight(handle->lterm);
	    
	    if(EqnIsEquLit(handle))
	    {
	       weight += TermWeight(handle->rterm, 0,VAR_FACTOR);
	       sweight += TermStandardWeight(handle->rterm);
	    }
	    if(sweight <= weight)
	    {
	       found = true;
	       break;
	    }
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 SelectPosOptimalLiteral(clause);
      }
      else
      {
	 EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}



/*-----------------------------------------------------------------------
//
// Function: SelectAllCondOptimalLiteral()
//
//   As above, but if the clause has only positive literals that are very
//   uninstantiated, select no literal at all.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectAllCondOptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool found = true;
      long weight, sweight;

      while(handle)
      {
	 if(EqnIsPositive(handle))
	 {
	    weight = TermWeight(handle->lterm, 0,VAR_FACTOR);
	    sweight = TermStandardWeight(handle->lterm);
	    
	    if(EqnIsEquLit(handle))
	    {
	       weight += TermWeight(handle->rterm, 0,VAR_FACTOR);
	       sweight += TermStandardWeight(handle->rterm);
	    }
	    if(sweight > weight)
	    {
	       found = false;
	       break;
	    }
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 SelectOptimalLiteral(clause);
      }
      else
      {
      EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPosAllCondOptimalLiteral()
//
//   As above, but select positive literals as well.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectPosAllCondOptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool found = true;
      long weight, sweight;

      while(handle)
      {
	 if(EqnIsPositive(handle))
	 {
	    weight = TermWeight(handle->lterm, 0,VAR_FACTOR);
	    sweight = TermStandardWeight(handle->lterm);
	    
	    if(EqnIsEquLit(handle))
	    {
	       weight += TermWeight(handle->rterm, 0,VAR_FACTOR);
	       sweight += TermStandardWeight(handle->rterm);
	    }
	    if(sweight > weight)
	    {
	       found = false;
	       break;
	    }
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 SelectPosOptimalLiteral(clause);
      }
      else
      {
	 EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}




/*-----------------------------------------------------------------------
//
// Function: SelectDepth2OptimalLiteral()
//
//   Select optimal literal unless there is a literal with depth <= 2,
//   then select no literal.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectDepth2OptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool found = false;

      while(handle)
      {
	 if(EqnDepth(handle)<=2)
	 {
	    found = true;
	    break;
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 SelectOptimalLiteral(clause);
      }
      else
      {
	 EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPosDepth2OptimalLiteral()
//
//   As above, but select positive literals as well.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectPosDepth2OptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool found = false;

      while(handle)
      {
	 if(EqnDepth(handle)<=2)
	 {
	    found = true;
	    break;
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 SelectPosOptimalLiteral(clause);
      }
      else
      {
	 EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPDepth2OptimalLiteral()
//
//   Select optimal literal unless there is a positive literal with
//   depth <= 2, then select no literal.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectPDepth2OptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool found = false;

      while(handle)
      {
	 if(EqnIsPositive(handle) && (EqnDepth(handle)<=2))
	 {
	    found = true;
	    break;
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 SelectOptimalLiteral(clause);
      }
      else
      {
	 EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPosPDepth2OptimalLiteral()
//
//   As above, with positive literals.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectPosPDepth2OptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool found = false;

      while(handle)
      {
	 if(EqnIsPositive(handle) && (EqnDepth(handle)<=2))
	 {
	    found = true;
	    break;
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 SelectPosOptimalLiteral(clause);
      }
      else
      {
	 EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}



/*-----------------------------------------------------------------------
//
// Function: SelectNDepth2OptimalLiteral()
//
//   Select optimal literal unless there is a negative literal with
//   depth <= 2, then select no literal.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectNDepth2OptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool found = false;

      while(handle)
      {
	 if(EqnIsNegative(handle) && (EqnDepth(handle)<=2))
	 {
	    found = true;
	    break;
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 SelectOptimalLiteral(clause);
      }
      else
      {
	 EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPosNDepth2OptimalLiteral()
//
//   As above, with positive literals.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void SelectPosNDepth2OptimalLiteral(Clause_p clause)
{
   if(clause->neg_lit_no)
   {
      Eqn_p handle = clause->literals;
      bool found = false;

      while(handle)
      {
	 if(EqnIsNegative(handle) && (EqnDepth(handle)<=2))
	 {
	    found = true;
	    break;
	 }
	 handle = handle->next;
      }
      if(!found)
      {
	 SelectPosOptimalLiteral(clause);
      }
      else
      {
	 EqnListDelProp(clause->literals, EPIsSelected);
      }
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}

/*-----------------------------------------------------------------------
//
// Function: SelectNonRROptimalLiteral()
//
//   If a clause is not range-restricted, select the optimal literal,
//   otherwise select no literal.
//
// Global Variables: -
//
// Side Effects    : Minimal ;-) (only by called functions)
//
/----------------------------------------------------------------------*/

void SelectNonRROptimalLiteral(Clause_p clause)
{
   if(!ClauseIsRangeRestricted(clause))
   {
      SelectOptimalLiteral(clause);
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*-----------------------------------------------------------------------
//
// Function: SelectPosNonRROptimalLiteral()
//
//   If a clause is not range-restricted, select the optimal literal
//   and positive literals, otherwise select no literal.
//
// Global Variables: -
//
// Side Effects    : Minimal ;-) (only by called functions)
//
/----------------------------------------------------------------------*/

void SelectPosNonRROptimalLiteral(Clause_p clause)
{
   if(!ClauseIsRangeRestricted(clause))
   {
      SelectPosOptimalLiteral(clause);
   }
   else
   {
      EqnListDelProp(clause->literals, EPIsSelected);
   }
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/

